from .decorator import Resolver, Query, Mutation, Subscription
from .graphql_module import GraphqlModule, GraphqlOption
from .pubsub import PubSub

__all__ = [
    "Resolver",
    "Query",
    "Mutation",
    "Subscription",
    "GraphqlModule",
    "GraphqlOption",
    "PubSub"
]



import asyncio
from urllib.parse import urlparse, parse_qs

from nestipy.core.nestipy_application import NestipyApplication
from .helpers import HeadersType, QueryType, CookiesType, get_example_scope, TestResponse


class TestSimulator:
    __test__ = False

    def __init__(self, app: NestipyApplication):
        self.app = app
        self.scope: dict = {}
        self.receive_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()

    @classmethod
    def _create_scope(
            cls,
            method: str,
            path: str,
            headers: HeadersType = None,
            query: QueryType = None,
            cookies: CookiesType = None,
    ) -> dict:
        """Creates a mocked ASGI scope"""
        return get_example_scope(
            method, path, extra_headers=headers, query=query, cookies=cookies
        )

    async def _simulate_receive(self):
        return await self.receive_queue.get()

    async def _simulate_send(self, message: dict):
        await self.send_queue.put(message)

    @classmethod
    def _extract_query_params(cls, path) -> QueryType:
        parsed_url = urlparse(path)
        return parse_qs(parsed_url.query)

    async def make_request(self, method: str, path: str, headers: dict[str, str] = None, body=b''):
        if headers is not None:
            headers = [(key.encode('utf-8'), value.encode('utf-8')) for key, value in headers]
        # app request
        self.scope = self._create_scope(method, urlparse(path).path, headers, query=self._extract_query_params(path))
        await self.receive_queue.put({
            'type': 'http.request',
            'body': body,
            'more_body': False,
        })
        await self.app.setup()
        await self.app(self.scope, self._simulate_receive, self._simulate_send)
        response = []
        while not self.send_queue.empty():
            response.append(await self.send_queue.get())
        return TestResponse.from_dict(response)



"""
This module provides common functions to handle Markdown.
These functions apply to any kind of Markdown work.
"""
from typing import Dict, Iterable


def write_row(
    row: Iterable[str],
    columns_widths: Dict[int, int],
    padding: int = 1,
    indent: int = 0,
) -> str:
    """
    Writes a single row for a Markdown table.
    """
    indent_chars = " " * indent
    pad_chars = " " * padding
    return (
        indent_chars
        + "|"
        + "|".join(
            pad_chars + str(cell_value).ljust(columns_widths[i]) + pad_chars
            for i, cell_value in enumerate(row)
        )
        + "|"
    )


def write_table_lines(
    matrix: Iterable[Iterable[str]],
    write_headers: bool = True,
    padding: int = 1,
    indent: int = 0,
) -> Iterable[str]:
    """
    Writes the lines of a Markdown table from a matrix (iterable of string records).
    """
    # TODO: assert that all rows have the same number of cells
    columns_widths = {
        i: max(len(str(value)) for value in column)
        for i, column in enumerate(zip(*matrix))
    }

    for row in matrix:
        yield write_row(row, columns_widths, padding, indent)

        if write_headers:
            # add separator line after headers
            yield write_row(
                ["-" * column_len for column_len in columns_widths.values()],
                columns_widths,
                padding,
                indent,
            )

            write_headers = False


def write_table(
    matrix: Iterable[Iterable[str]],
    write_headers: bool = True,
    padding: int = 1,
) -> str:
    """
    Writes a Markdown table from a matrix (iterable of string records).
    """
    return "\n".join((write_table_lines(matrix, write_headers, padding)))


def normalize_link(value: str) -> str:
    if not value:
        raise ValueError("Missing value")
    return value.replace(".", "")



from .request import Request
from .response import Response
from .websocket import Websocket
from .upload_file import UploadFile

__all__ = [
    "Response",
    "Request",
    "Websocket",
    "UploadFile"
]



from strawberry import directive as Directive, schema_directive as SchemaDirective
from strawberry import field as Field, lazy as Lazy, parent as Parent, argument as Argument
from strawberry import input as Input, scalar as Scalar, enum as Enum
from strawberry import type as ObjectType, interface as Interface

from .strawberry_asgi import StrawberryAsgi
from .strawberry_adapter import StrawberryAdapter

__all__ = [
    "StrawberryAsgi",
    "StrawberryAdapter",
    "ObjectType",
    "Interface",
    "Input",
    "Scalar",
    "Directive",
    "SchemaDirective",
    "Enum",
    "Field",
    "Lazy",
    "Lazy",
    "Parent",
    "Argument"
]



from nestipy.core.nestipy_application import NestipyApplication
from .helpers import TestResponse
from .simulator import TestSimulator


class TestClient:
    __test__ = False

    def __init__(self, app: NestipyApplication):
        self._simulator = TestSimulator(app)

    async def get(self, path: str, headers: dict = None, body: bytes = b'') -> TestResponse:
        return await self._simulator.make_request('GET', path, headers, body)

    async def post(self, path: str, headers: dict = None, body: bytes = b'') -> TestResponse:
        return await self._simulator.make_request('POST', path, headers, body)

    async def put(self, path: str, headers: dict = None, body: bytes = b'') -> TestResponse:
        return await self._simulator.make_request('PUT', path, headers, body)

    async def delete(self, path: str, headers: dict = None, body: bytes = b'') -> TestResponse:
        return await self._simulator.make_request('DELETE', path, headers, body)

    async def options(self, path: str, headers: dict = None, body: bytes = b'') -> TestResponse:
        return await self._simulator.make_request('OPTIONS', path, headers, body)



from abc import ABC, abstractmethod
from typing import Any

import ujson


class Serializer(ABC):

    @abstractmethod
    async def serialize(self, data: Any) -> str:
        pass

    @abstractmethod
    async def deserialize(self, data: str) -> Any:
        pass


class JSONSerializer(Serializer):
    async def serialize(self, data: Any) -> str:
        return ujson.dumps(data)

    async def deserialize(self, data: str) -> Any:
        return ujson.loads(data)



from .cors import cors
from .helmet import helmet
from .interface import NestipyMiddleware
from .meta import MiddlewareKey
from .session import session, SessionOption

__all__ = [
    "cors",
    "NestipyMiddleware",
    "MiddlewareKey",
    "SessionOption",
    "session",
    "helmet"
]



from abc import ABC, abstractmethod
from typing import Type, Union

from nestipy.common.utils import uniq_list
from nestipy.dynamic_module import DynamicModule
from nestipy.ioc import NestipyContainer
from nestipy.metadata import ClassMetadata, ModuleMetadata, Reflect


class MetadataCreator(ABC):

    def __init__(self, module, global_data=None, is_root=False):
        self.module = self._dynamic_module_to_module(module)
        self.is_root = is_root
        self.global_data = global_data or []

    @abstractmethod
    def _extract(self) -> list:
        pass

    @abstractmethod
    def _type(self) -> Type["MetadataCreator"]:
        pass

    def extract_providers(self) -> list:
        return uniq_list(Reflect.get_metadata(self.module, ModuleMetadata.Providers, []))

    def _put_dependency_metadata(self) -> None:
        data = self._extract()
        for p in data:
            global_data = self.extract_providers() if self.is_root else self.global_data
            # only p that have not metadata
            if not hasattr(p, ClassMetadata.Metadata):
                Reflect.set_metadata(
                    p,
                    ClassMetadata.Metadata,
                    ClassMetadata(
                        self.module,
                        global_providers=global_data
                    )
                )

    def _extract_import(self) -> list:
        """
        Extract module imported in Module
        :return: list
        """
        return uniq_list(
            [self._dynamic_module_to_module(m) for m in Reflect.get_metadata(self.module, ModuleMetadata.Imports, [])])

    @classmethod
    def _dynamic_module_to_module(cls, im: Union[Type, object]):
        if isinstance(im, DynamicModule):
            Reflect.set_metadata(im.module, ModuleMetadata.Providers, im.providers)
            Reflect.set_metadata(im.module, ModuleMetadata.Controllers, im.controllers)
            Reflect.set_metadata(im.module, ModuleMetadata.Imports, im.imports)
            Reflect.set_metadata(im.module, ModuleMetadata.Exports, im.exports)
            Reflect.set_metadata(im.module, ModuleMetadata.Global, im.is_global)
            NestipyContainer.get_instance().add_singleton(im.module)
            return im.module
        else:
            return im

    def create(self) -> None:
        """
        Compile is about put module parent to the dependency metadata of a provider or controller
        :return:
        """
        if self.is_root:
            Reflect.set_metadata(self.module, ModuleMetadata.Root, True)
            self.global_data = self.extract_providers()
            # compile global first
            imports = self._extract_import()
            not_global_module = []
            for im in imports:
                if not Reflect.get_metadata(im, ModuleMetadata.Global, False):
                    not_global_module.append(im)
                else:
                    self._type()(im, global_data=self.global_data).create()
                    import_providers_form_exports = []
                    for export in Reflect.get_metadata(im, ModuleMetadata.Exports, []):
                        # For module re-exporting
                        is_module: bool = Reflect.get_metadata(export, ModuleMetadata.Module, False)
                        if is_module:
                            import_providers_form_exports = import_providers_form_exports + Reflect.get_metadata(
                                export,
                                ModuleMetadata.Providers,
                                []
                            )
                        else:
                            import_providers_form_exports.append(export)
                    self.global_data = self.global_data + import_providers_form_exports

            # compile non global after
            for im in not_global_module:
                self._type()(im, global_data=self.global_data).create()
            self._put_dependency_metadata()
        else:
            self._put_dependency_metadata()



from .class_ import ClassMetadata
from .container import NestipyContextProperty
from .decorator import SetMetadata
from .dependency import CtxDepKey
from .module import ModuleMetadata
from .provider_token import ProviderToken
from .reflect import Reflect
from .route import RouteKey

__all__ = [
    "ProviderToken",
    "ClassMetadata",
    "NestipyContextProperty",
    "ModuleMetadata",
    "SetMetadata",
    "CtxDepKey",
    "RouteKey",
    "Reflect"
]



from typing import (
    Union, Callable,
)

from starlette.requests import Request
from starlette.responses import HTMLResponse, PlainTextResponse, Response
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.http.exceptions import HTTPException
from strawberry.schema import BaseSchema

from ..graphql_asgi import GraphqlAsgi
from ..graphql_module import GraphqlOption


class StrawberryAsgi(GraphQL, GraphqlAsgi):

    def __init__(self, schema: BaseSchema, option: GraphqlOption):
        super().__init__(schema)
        self.set_graphql_option(option)

    async def render_graphql_ide(self, request: Union[Request, WebSocket]) -> Response:
        if self.option.ide:
            return HTMLResponse(await self.get_graphql_ide())
        else:
            raise HTTPException(404, "Not Found")

    async def handle(self, scope: dict, receive, send):
        await super().handle(scope, receive, send)
        await self.__call__(scope, receive, send)

    async def handle_http(
            self,
            scope: dict,
            receive: Callable,
            send: Callable,
    ) -> None:
        request = Request(scope=scope, receive=receive)
        try:
            response = await self.run(request)
        except HTTPException as e:
            response = PlainTextResponse(e.reason, status_code=e.status_code)
        # Apply cors to graphql
        response.headers.update({
            'access-control-allow-origin': '*',
            'access-control-allow-headers': 'Content-Type',
            'access-control-allow-methods': 'GET, POST, PUT, DELETE, OPTIONS'
        })
        await response(scope, receive, send)



from abc import ABC

from .consumer import MiddlewareConsumer


class NestipyModule(ABC):
    def configure(self, consumer: MiddlewareConsumer):
        pass

    async def on_startup(self):
        pass

    async def on_shutdown(self):
        pass



from typing import Callable, Any

from nestipy.metadata import Reflect
from .meta import TemplateKey


def Render(template: str) -> Callable[[Callable], Any]:
    """
    Args:
        template(str):

    Returns:
        callable(Callable): A callable decorator
    """

    def wrapper(method: Callable) -> Callable:
        Reflect.set_metadata(method, TemplateKey.MetaRender, template)
        return method

    return wrapper



from typing import Any, cast

from nestipy.dynamic_module import ConfigurableModuleBuilder, DynamicModule
from dataclasses import dataclass


@dataclass
class EventOption:
    pass


def extract_callback(dynamic_module: DynamicModule, extras: dict[str, Any]):
    if extras.get('is_global'):
        dynamic_module.is_global = cast(bool, extras.get('is_global'))


ConfigurableClassBuilder, EVENT_CONFIG = ConfigurableModuleBuilder[EventOption]().set_extras(
    {},
    extract_callback
).build()



import dataclasses
import os.path
import traceback
from typing import Type, Callable, Literal, Union, Any, TYPE_CHECKING, Optional, Dict

from openapidocs.v3 import PathItem, Schema, Reference
from rich.traceback import install

from nestipy.common.http_ import Response, Request
from nestipy.common.logger import logger, console
from nestipy.common.middleware import NestipyMiddleware
from nestipy.common.template import TemplateEngine, TemplateKey
from nestipy.common.utils import uniq_list
from nestipy.core.template import MinimalJinjaTemplateEngine
from nestipy.dynamic_module import DynamicModule
from nestipy.graphql.graphql_adapter import GraphqlAdapter
from nestipy.graphql.strawberry.strawberry_adapter import StrawberryAdapter
from nestipy.ioc import MiddlewareContainer, MiddlewareProxy, NestipyContainer, ModuleProviderDict
from nestipy.metadata import ModuleMetadata, Reflect
from .adapter.fastapi_adapter import FastApiAdapter
from .adapter.http_adapter import HttpAdapter
from .discover import DiscoverService
from .instance_loader import InstanceLoader
from .meta.controller_metadata_creator import ControllerMetadataCreator
from .meta.module_metadata_creator import ModuleMetadataCreator
from .meta.provider_metadata_creator import ProviderMetadataCreator
from .router.router_proxy import RouterProxy
from ..graphql.graphql_proxy import GraphqlProxy
from ..types_ import NextFn
from ..websocket.adapter import IoAdapter
from ..websocket.proxy import IoSocketProxy

if TYPE_CHECKING:
    from nestipy.common.exception.interface import ExceptionFilter
    from nestipy.common.interceptor import NestipyInterceptor
    from nestipy.common.guards.can_activate import CanActivate

# install rich track_back
install(console=console, width=200)


@dataclasses.dataclass
class NestipyConfig:
    adapter: HttpAdapter = None
    cors: bool = None
    debug: bool = True


class NestipyApplication:
    _root_module: Type = None
    _openapi_paths = {}
    _openapi_schemas = {}
    _prefix: Union[str | None] = None
    _debug: bool = True
    _ready: bool = False

    def __init__(self, config: NestipyConfig = None):
        config = config if config is not None else NestipyConfig()
        # self._http_adapter: HttpServer = FastAPIAdapter()
        self._http_adapter: HttpAdapter = config.adapter if config.adapter is not None else FastApiAdapter()
        self._graphql_builder = StrawberryAdapter()
        self._router_proxy = RouterProxy(self._http_adapter)
        self._middleware_container = MiddlewareContainer.get_instance()
        self.instance_loader = InstanceLoader()
        self.process_config(config)
        self.on_shutdown(self._destroy)

    def on_startup(self, callback: Callable):
        self._http_adapter.on_startup_callback(callback)
        return callback

    def on_shutdown(self, callback: Callable):
        self._http_adapter.on_shutdown_callback(callback)
        return callback

    @classmethod
    async def get(cls, key: Union[Type, str]):
        return await NestipyContainer.get_instance().get(key)

    def process_config(self, config: NestipyConfig):
        if config.cors is not None:
            self._http_adapter.enable_cors()
        self._debug = config.debug
        self._http_adapter.debug = config.debug

    @classmethod
    def _get_modules(cls, module: Type) -> list[Type]:
        modules: list[Type] = [module]
        for m in Reflect.get_metadata(module, ModuleMetadata.Imports, []):
            if isinstance(m, DynamicModule):
                modules.append(m.module)
            else:
                modules.append(m)
        return uniq_list(modules)

    def init(self, root_module: Type):
        self._root_module = root_module
        self._add_root_module_provider(DiscoverService, _init=False)
        self._set_metadata()

    def _set_metadata(self):
        provider_metadata_maker = ProviderMetadataCreator(self._root_module, is_root=True)
        provider_metadata_maker.create()

        controller_metadata_maker = ControllerMetadataCreator(self._root_module, is_root=True)
        controller_metadata_maker.create()

        # optional
        module_metadata_maker = ModuleMetadataCreator(self._root_module)
        module_metadata_maker.create()

    async def setup(self):
        try:
            modules = self._get_modules(self._root_module)
            graphql_module_instance = await self.instance_loader.create_instances(modules)
            # create and register route to platform adapter
            self._openapi_paths, self._openapi_schemas = self._router_proxy.apply_routes(modules, self._prefix)
            # check if graphql is enabled
            if graphql_module_instance is not None:
                GraphqlProxy(self._http_adapter, self._graphql_builder).apply_resolvers(
                    graphql_module_instance,
                    modules
                )
            if self._http_adapter.get_io_adapter() is not None:
                IoSocketProxy(self._http_adapter).apply_routes(modules)
            # Register open api catch asynchronously
            from nestipy.openapi.constant import OPENAPI_HANDLER_METADATA
            openapi_register: Callable = Reflect.get_metadata(self, OPENAPI_HANDLER_METADATA, None)
            if openapi_register is not None:
                openapi_register()

            self._ready = True

        except Exception as e:
            _tb = traceback.format_exc()
            logger.error(e)
            logger.error(_tb)
        finally:
            # Register devtools static path
            self.use_static_assets(
                os.path.realpath(os.path.join(os.path.dirname(__file__), '..', 'devtools', 'frontend', 'static')),
                '_devtools/static'
            )
            if self._debug:
                # Not found
                not_found_path = self._http_adapter.create_wichard().lstrip('/')
                self._http_adapter.get(not_found_path, self._router_proxy.render_not_found, {})

    async def ready(self) -> bool:
        if not self._ready:
            await self.setup()
        return self._ready

    async def _destroy(self):
        await self.instance_loader.destroy()

    def get_adapter(self) -> HttpAdapter:
        return self._http_adapter

    def get_graphql_adapter(self) -> GraphqlAdapter:
        return self._graphql_builder

    def get_openapi_paths(self) -> dict[Any, PathItem]:
        return self._openapi_paths

    def get_open_api_schemas(self) -> Optional[Dict[str, Union[Schema, Reference]]]:
        return self._openapi_schemas

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        if scope.get('type') == 'lifespan':
            await self.ready()
        await self.get_adapter()(scope, receive, send)

    def use(self, *middleware: Union[Type[NestipyMiddleware], Callable]):
        for m in middleware:
            if issubclass(m, NestipyMiddleware) or callable(m):
                proxy = MiddlewareProxy(m)
                self._middleware_container.add_singleton(proxy)
        self._add_root_module_provider(*middleware)

    def enable_cors(self):
        self._http_adapter.enable_cors()

    def use_static_assets(self, assets_path: str, url: str = '/static'):
        async def render_asset_file(req: "Request", res: "Response", _next_fn: "NextFn") -> Response:
            file_path = os.path.join(assets_path, req.path.replace(f'/{url.strip("/")}', '').strip('/'))
            return await res.download(
                file_path=file_path,
                attachment=False
            )

        static_path = self._http_adapter.create_wichard(f'/{url.strip("/")}')
        self._http_adapter.get(static_path, render_asset_file, {})

    def set_base_view_dir(self, view_dir: str):
        self._setup_template_engine(view_dir)

    def set_view_engine(self, engine: Union[Literal['minijinja'], TemplateEngine]):
        if isinstance(engine, TemplateEngine):
            self._http_adapter.set(TemplateKey.MetaEngine, engine)
        else:
            # TODO, add more choice template like jinja2
            pass

    def _setup_template_engine(self, view_dir):
        engine = MinimalJinjaTemplateEngine(template_dir=view_dir)
        self._http_adapter.set(TemplateKey.MetaEngine, engine)

    def get_template_engine(self) -> Union[TemplateEngine, None]:
        engine: Union[TemplateEngine, None] = self._http_adapter.get_state(TemplateKey.MetaEngine)
        if engine is None:
            raise Exception('Template engine not configured')
        return engine

    def use_global_interceptors(self, *interceptors: Union[Type, "NestipyInterceptor"]):
        self._http_adapter.add_global_interceptors(*interceptors)
        self._add_root_module_provider(*interceptors)

    def use_global_filters(self, *filters: Union[Type, "ExceptionFilter"]):
        self._http_adapter.add_global_filters(*filters)
        self._add_root_module_provider(*filters)

    def use_global_guards(self, *guards: Union[Type, "CanActivate"]):
        self._http_adapter.add_global_guards(*guards)
        # self._add_root_module_provider(*guards)

    def use_global_prefix(self, prefix: str = ""):
        self._prefix = prefix or ""

    def _add_root_module_provider(self, *providers: Union[ModuleProviderDict, Type], _init: bool = True):
        root_providers: list = Reflect.get_metadata(self._root_module, ModuleMetadata.Providers, [])
        root_providers = root_providers + list(providers)
        Reflect.set_metadata(self._root_module, ModuleMetadata.Providers, root_providers)
        #  re init setting metadata
        if _init:
            self._set_metadata()

    def add_module_root_module(self, *modules: Type, _init: bool = True):
        root_imports: list = Reflect.get_metadata(self._root_module, ModuleMetadata.Imports, [])
        root_imports = root_imports + list(modules)
        Reflect.set_metadata(self._root_module, ModuleMetadata.Imports, root_imports)
        #  re init setting metadata
        if _init:
            self._set_metadata()

    def use_io_adapter(self, io_adapter: IoAdapter):
        # edit root module provider
        # TODO refactor
        NestipyContainer.get_instance().add_singleton_instance(IoAdapter, io_adapter)
        self._add_root_module_provider(ModuleProviderDict(token=IoAdapter, value=io_adapter))
        # setup io adapter to http_adapter
        self._http_adapter.use_io_adapter(io_adapter=io_adapter)



import asyncio

from nestipy.microservice.client import NatsClientProxy, MicroserviceOption


async def main():
    client = NatsClientProxy(
        MicroserviceOption(
            url="nats://localhost:4222"
        ),
    )
    await client.send("topic", "request")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from typing import AsyncIterator

import nats
from nats.aio.client import Client
from nats.aio.subscription import Subscription

from .base import ClientProxy, MicroserviceOption

NATS_QUEUE = "microservice:queue"


class NatsClientProxy(ClientProxy):
    client: Client
    consumer: Subscription

    def __init__(self, option: MicroserviceOption):
        super().__init__(option)

    async def slave(self) -> "ClientProxy":
        return NatsClientProxy(
            self.option,
        )

    async def connect(self):
        self.client = await nats.connect(self.option.url)

    async def _publish(self, topic: str, data: str):
        await self.client.publish(topic, data.encode())

    async def subscribe(self, *args, **kwargs):
        self.consumer = await self.client.subscribe(*args, **kwargs, queue=NATS_QUEUE)

    async def unsubscribe(self, *args):
        await self.consumer.unsubscribe()

    async def send_response(self, topic: str, data: str):
        await self._publish(topic, data)

    async def listen(self) -> AsyncIterator[str]:
        while True:
            async for msg in self.consumer.messages:
                yield msg.data.decode('utf-8')
            await asyncio.sleep(0.01)

    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        while True:
            async for msg in self.consumer.messages:
                return msg.data.decode('utf-8')
            await asyncio.sleep(0.01)

    async def close(self):
        await self.client.close()



from unittest.mock import MagicMock, patch, Mock

import pytest

from nestipy.common.decorator.class_ import Controller, Module, ModuleMetadata, RouteKey


# Mocks for NestipyContainer and Reflect
@pytest.fixture
def mock_container() -> Mock:
    with patch('nestipy.ioc.container.NestipyContainer') as MockContainer:
        instance = MockContainer.get_instance.return_value
        yield instance


@pytest.fixture
def mock_reflect() -> Mock:
    with patch('nestipy.metadata.Reflect') as MockReflect:
        yield MockReflect


# Controller tests
def test_controller(mock_container: Mock, mock_reflect: Mock) -> None:
    @Controller(path='/test')
    class TestClass:
        pass

    mock_container.add_singleton.assert_called_once_with(TestClass)
    mock_reflect.set_metadata.assert_any_call(TestClass, RouteKey.path, '/test')
    mock_reflect.set_metadata.assert_any_call(TestClass, RouteKey.kwargs, {})


# Module tests
def test_module(mock_container, mock_reflect):
    providers = [MagicMock()]
    controllers = [MagicMock()]
    imports = [MagicMock()]
    exports = [MagicMock()]
    is_global = True

    @Module(providers=providers, controllers=controllers, imports=imports, exports=exports, is_global=is_global)
    class TestModule:
        pass

    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Providers, providers)
    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Controllers, controllers)
    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Imports, imports)
    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Exports, exports)
    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Global, is_global)
    mock_reflect.set_metadata.assert_any_call(TestModule, ModuleMetadata.Module, True)
    mock_container.add_singleton.assert_called_once_with(TestModule)



from typing import get_args, Any, Type, cast

from .dependency import TypeAnnotated, Instance


class ContainerHelper:

    @classmethod
    def get_type_from_annotation(cls, annotation: Any) -> tuple[Type, TypeAnnotated]:
        args: tuple = get_args(annotation)
        if len(args) == 2:
            return cast(tuple[Type, TypeAnnotated], args)
        else:
            return annotation, Instance()

    @classmethod
    def get_value_from_dict(cls, values: dict, key: str, default=None):
        if key in values.keys():
            return values[key]
        else:
            return default



import asyncio
from typing import AsyncIterator, Union, cast

import async_timeout
from aio_pika import connect_robust, Message, ExchangeType
from aio_pika.abc import AbstractRobustConnection, AbstractRobustChannel
from aio_pika.abc import AbstractRobustQueue, AbstractRobustExchange

from .base import ClientProxy, MicroserviceOption


class RabbitMQClientProxy(ClientProxy):
    connection: AbstractRobustConnection
    channel: AbstractRobustChannel
    exchange: AbstractRobustExchange
    consumer_queue: Union[AbstractRobustQueue, None]
    CHANGE = "microservice:exchange"

    def __init__(self, option: MicroserviceOption):
        super().__init__(option)

    async def slave(self) -> "ClientProxy":
        return RabbitMQClientProxy(
            self.option,
        )

    async def connect(self):
        self.connection = await connect_robust(
            host=self.option.option.host,
            port=self.option.option.port
        )
        self.channel = await self.connection.channel()
        await self.channel.__aenter__()
        self.exchange = await self.channel.declare_exchange(
            self.CHANGE,
            ExchangeType.FANOUT
        )

    async def _publish(self, topic: str, data: str):
        await self.exchange.publish(
            Message(
                body=data.encode()
            ),
            routing_key=topic
        )

    async def send_response(self, topic: str, data: str):
        response_exchange = await self.channel.declare_exchange(
            f"{self.CHANGE}:{topic}",
            ExchangeType.FANOUT
        )
        await response_exchange.publish(
            Message(
                body=data.encode()
            ),
            routing_key=topic
        )

    async def subscribe(self, *args, **kwargs):
        self.consumer_queue = await self.channel.declare_queue(*args)
        await self.consumer_queue.bind(self.exchange)

    async def unsubscribe(self, *args):
        await self.consumer_queue.unbind(self.exchange)
        await self.consumer_queue.delete()

    async def listen(self) -> AsyncIterator[str]:
        if self.consumer_queue is not None:
            async with self.consumer_queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        msg = cast(Message, message)
                        yield msg.body.decode('utf-8')

    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        response_exchange = await self.channel.declare_exchange(
            f"{self.CHANGE}:{from_topic}",
            ExchangeType.FANOUT
        )
        response_queue = await self.channel.declare_queue(from_topic)
        await response_queue.bind(response_exchange)
        try:
            with async_timeout.timeout(timeout):
                async with response_queue.iterator() as queue_iter:
                    async for message in queue_iter:
                        async with message.process():
                            msg = cast(Message, message)
                            if message.routing_key == from_topic:
                                res = msg.body.decode('utf-8')
                                break

        except asyncio.TimeoutError as e:
            raise e
        finally:
            await response_queue.unbind(response_exchange)
            await response_queue.delete()
            await response_exchange.delete()
        return res

    async def close(self):
        await self.channel.__aexit__(None, None, None)
        await self.connection.close()



from dataclasses import is_dataclass
from typing import Any, Type, Optional

from pydantic import BaseModel

from nestipy.metadata import CtxDepKey
from .annotation import ParamAnnotation, TypeAnnotatedCallable
from .context_container import RequestContextContainer


class TypeAnnotated:

    def __init__(self, metadata: ParamAnnotation):
        self.metadata = metadata

    def __call__(self, token: Any = None) -> "TypeAnnotated":
        return TypeAnnotated(ParamAnnotation(self.metadata.callback, self.metadata.key, token))


def create_type_annotated(callback: TypeAnnotatedCallable, key: str) -> TypeAnnotated:
    return TypeAnnotated(ParamAnnotation(callback, key))


def inject_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return None


def instance_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return None


def req_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _request_context.execution_context.get_request()


def res_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _request_context.execution_context.get_response()


def to_valid_value(value: Any, _type_ref: Type):
    if is_dataclass(_type_ref):
        return _type_ref(**value)
    elif issubclass(_type_ref, BaseModel):
        return _type_ref(**value)
    return value


async def body_callback(
        _name: str,
        _token: Optional[str],
        _type_ref: Type,
        _request_context: RequestContextContainer
):
    req = _request_context.execution_context.get_request()
    form_data = await req.form()
    if bool(form_data):
        return to_valid_value(form_data, _type_ref)
    else:
        return to_valid_value(await req.json(), _type_ref)


def _get_request_param_value(
        key: str,
        _type_ref: Type, _request_context: RequestContextContainer,
        token: Optional[str] = None
):
    req = _request_context.execution_context.get_request()
    value: dict = getattr(req, key)
    if token:
        return value.get(token)
    else:
        return to_valid_value(value, _type_ref)


def session_callback(_name: str, token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _get_request_param_value('session', _type_ref, _request_context, token)


def cookie_callback(_name: str, token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _get_request_param_value('cookies', _type_ref, _request_context, token)


def query_callback(_name: str, token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _get_request_param_value('query_params', _type_ref, _request_context, token)


def params_callback(_name: str, token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _get_request_param_value('path_params', _type_ref, _request_context, token)


def headers_callback(_name: str, token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _get_request_param_value('headers', _type_ref, _request_context, token)


def args_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    args = _request_context.execution_context.switch_to_graphql().get_args()
    return args.get(_name)


def context_callback(_name: str, _token: Optional[str], _type_ref: Type, _request_context: RequestContextContainer):
    return _request_context


def graphql_context_callback(_name: str, _token: Optional[str], _type_ref: Type,
                             _request_context: RequestContextContainer):
    return _request_context.execution_context.switch_to_graphql().get_context()


def websocket_server_callback(_name: str, _token: Optional[str], _type_ref: Type,
                              _request_context: RequestContextContainer):
    return _request_context.execution_context.switch_to_websocket().get_server()


def websocket_client_callback(_name: str, _token: Optional[str], _type_ref: Type,
                              _request_context: RequestContextContainer):
    return _request_context.execution_context.switch_to_websocket().get_client()


def websocket_data_callback(_name: str, _token: Optional[str], _type_ref: Type,
                            _request_context: RequestContextContainer):
    return _request_context.execution_context.switch_to_websocket().get_data()


Instance = create_type_annotated(instance_callback, 'instance')
Inject = create_type_annotated(inject_callback, CtxDepKey.Service)
Req = create_type_annotated(req_callback, CtxDepKey.Request)
Res = create_type_annotated(res_callback, CtxDepKey.Response)
Session = create_type_annotated(session_callback, CtxDepKey.Session)
Cookie = create_type_annotated(cookie_callback, CtxDepKey.Cookie)
Query = create_type_annotated(query_callback, CtxDepKey.Query)
Body = create_type_annotated(body_callback, CtxDepKey.Body)
Param = create_type_annotated(params_callback, CtxDepKey.Param)
Arg = create_type_annotated(args_callback, CtxDepKey.Args)
Context = create_type_annotated(context_callback, CtxDepKey.Context)
GraphQlContext = create_type_annotated(graphql_context_callback, CtxDepKey.Context)
Header = create_type_annotated(headers_callback, CtxDepKey.Header)
WebSocketServer = create_type_annotated(websocket_server_callback, CtxDepKey.Service)
WebSocketClient = create_type_annotated(websocket_client_callback, CtxDepKey.WebSocketClient)
SocketData = create_type_annotated(websocket_data_callback, CtxDepKey.SocketData)



from abc import ABC, abstractmethod
from typing import Any


class TemplateEngine(ABC):

    def __init__(self, template_dir: str):
        self.template_dir = template_dir

    @abstractmethod
    def render(self, template: str, context: dict) -> str:
        """
        A specific implementation of rendering content of template file.

        Args:
            template (str): The template string.
            context (dict): The context data to be used in rendering.

        Returns:
            str: The rendered template.
        """
        pass

    @abstractmethod
    def render_str(self, string: str, context: dict) -> str:
        """
        Args:
            string (str): The string to compile.
            context (dict): The context data to be used in rendering.
        Returns:
            str: The rendered string .
        """
        pass

    @abstractmethod
    def get_env(self) -> Any:
        pass



from typing import Any

import httpx

http_client = httpx.Client(verify=False, timeout=20)


class FailedRequestError(Exception):
    def __init__(self, message) -> None:
        super().__init__(
            f"Failed request: {message}. "
            "Inspect the inner exception (__context__) for more information."
        )

    @property
    def inner_exception(self):
        return self.__context__


def ensure_success(response: httpx.Response) -> None:
    if response.status_code < 200 or response.status_code > 399:
        raise FailedRequestError(
            "Response status does not indicate success: "
            f"{response.status_code} {response.reason_phrase}"
        )


def http_get(url: str) -> Any:
    try:
        return http_client.get(url)
    except httpx.HTTPError as http_error:
        raise FailedRequestError(str(http_error)) from http_error



import asyncio
from typing import Any, AsyncIterator

from aiomqtt import Client

from .base import ClientProxy, MicroserviceOption


class MqttClientProxy(ClientProxy):
    client: Client

    def __init__(self, option: MicroserviceOption, ):
        super().__init__(option)
        self.client = Client(
            self.option.option.host,
            port=self.option.option.port,
            max_queued_incoming_messages=10,
            max_queued_outgoing_messages=10
        )

    async def slave(self) -> "ClientProxy":
        return MqttClientProxy(
            option=self.option,
        )

    async def connect(self):
        await self.client.__aenter__()

    async def _publish(self, topic, data):
        await self.client.publish(topic, data, retain=False)

    async def subscribe(self, *args, **kwargs):
        await self.client.subscribe(*args, **kwargs)
        pass

    async def unsubscribe(self, *args):
        await self.client.unsubscribe(*args)
        pass

    async def send_response(self, topic, data):
        await self._publish(topic, data)

    async def listen(self) -> AsyncIterator[str]:
        async for message in self.client.messages:
            yield message.payload.decode('utf-8')
        await asyncio.sleep(0.01)

    async def listen_response(self, from_topic: str, timeout: int = 30) -> Any:

        async for message in self.client.messages:
            return message.payload.decode('utf-8')
        await asyncio.sleep(0.01)

    async def close(self):
        await self.client.__aexit__(None, None, None)



import asyncio
import uuid
from abc import ABC, abstractmethod
from dataclasses import asdict
from dataclasses import field, dataclass
from typing import AsyncIterator
from typing import Optional

import async_timeout

from nestipy.microservice.context import RpcRequest, RpcResponse, MICROSERVICE_CHANNEL
from nestipy.microservice.exception import RpcException, RPCErrorMessage, RPCErrorCode
from nestipy.microservice.serilaizer import Serializer, JSONSerializer
from .option import Transport, MicroserviceClientOption


@dataclass
class MicroserviceOption:
    transport: Transport = field(default=Transport.REDIS)
    option: Optional[MicroserviceClientOption] = None
    url: Optional[str] = None
    proxy: Optional["ClientProxy"] = None
    serializer: Serializer = field(default=JSONSerializer())


class ClientProxy(ABC):

    def __init__(self, option: MicroserviceOption):
        self.option = option

    @abstractmethod
    async def slave(self) -> "ClientProxy":
        pass

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def _publish(self, topic, data):
        pass

    @abstractmethod
    async def subscribe(self, *args, **kwargs):
        pass

    @abstractmethod
    async def unsubscribe(self, *args):
        pass

    @abstractmethod
    async def send_response(self, topic, data):
        pass

    async def send(self, topic, data: any, headers: dict[str, str] = None) -> RpcResponse:
        await self.connect()
        response_topic = uuid.uuid4().hex
        request = RpcRequest(
            pattern=topic,
            data=data,
            response_topic=response_topic,
            headers=headers or {}
        )
        await self.subscribe(response_topic)
        await self._publish(
            MICROSERVICE_CHANNEL,
            await self.option.serializer.serialize(
                asdict(request)
            )
        )
        #  add timeout
        try:
            with async_timeout.timeout(60):
                rpc_response = await self.listen_response(response_topic, 60)
        except asyncio.TimeoutError:
            raise RpcException(
                status_code=RPCErrorCode.DEADLINE_EXCEEDED,
                message=RPCErrorMessage.DEADLINE_EXCEEDED
            )
        await self.unsubscribe(response_topic)
        await self.close()
        json_rep = await self.option.serializer.deserialize(rpc_response)
        response = RpcResponse.from_dict(json_rep)
        if response.exception:
            raise response.exception
        return response

    async def emit(self, topic, data, headers: dict[str, str] = None):
        request = RpcRequest(
            pattern=topic,
            data=data,
            headers=headers or {}
        )
        json_req = await self.option.serializer.serialize(
            asdict(request)
        )
        await self._publish(
            topic,
            json_req
        )

    @abstractmethod
    def listen(self) -> AsyncIterator[str]:
        pass

    @abstractmethod
    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        pass

    @abstractmethod
    async def close(self):
        pass



from typing import Union, Awaitable, Any, Callable, Type

from .container import NestipyContainer


class ModuleProviderDict:
    inject: list = []
    imports: list = []
    token: Union[str, Type]
    value: Any = None
    factory: Callable[..., Union[Awaitable, Any]] = None
    existing: Union[Type, str] = None
    use_class: Type = None

    def __init__(
            self,
            token: Union[str, Type],
            value: Any = None,
            factory: Callable[..., Union[Awaitable, Any]] = None,
            existing: Union[Type, str] = None,
            use_class: Type = None,
            inject: list = None,
    ):
        self.token = token
        self.value = value
        self.factory = factory
        self.existing = existing
        self.use_class = use_class
        self.inject = inject or []
        NestipyContainer.get_instance().add_singleton_instance(token, self)



from .base import ClientProxy, MicroserviceOption
from .mqtt import MqttClientProxy
from .nats import NatsClientProxy
from .option import Transport
from .rabbitmq import RabbitMQClientProxy
from .redis import RedisClientProxy


class ClientModuleFactory:
    @classmethod
    def create(cls, option: MicroserviceOption) -> ClientProxy:
        match option.transport:
            case Transport.REDIS:
                return RedisClientProxy(
                    option,
                )
            case Transport.MQTT:
                return MqttClientProxy(
                    option,
                )
            case Transport.RABBITMQ:
                return RabbitMQClientProxy(
                    option,
                )
            case Transport.NATS:
                return NatsClientProxy(
                    option,
                )
            case Transport.CUSTOM:
                if option.proxy is not None and isinstance(option.proxy, ClientProxy):
                    return option.proxy
                else:
                    raise Exception("Transport not known")
            case _:
                raise Exception("Transport not known")



import dataclasses
import os.path
from typing import TYPE_CHECKING

import aiofiles
import ujson

from nestipy.common.http_ import Request, Response
from nestipy.metadata import Reflect
from .constant import OPENAPI_HANDLER_METADATA
from .openapi_docs.common import Serializer
from .openapi_docs.v3 import OpenAPI

if TYPE_CHECKING:
    from ..core.nestipy_application import NestipyApplication


@dataclasses.dataclass
class _Document:
    json: str
    yml: str
    obj: dict


class SwaggerModule:

    @classmethod
    def remove_null(cls, json_data):
        if isinstance(json_data, dict):
            return {key: cls.remove_null(value) for key, value in json_data.items() if value is not None}
        elif isinstance(json_data, list):
            return [cls.remove_null(item) for item in json_data if item is not None]
        else:
            return json_data

    @classmethod
    def _create_document(cls, app: "NestipyApplication", config: OpenAPI) -> _Document:
        paths = app.get_openapi_paths()
        config.paths = paths
        schemas = app.get_open_api_schemas()
        if "schemas" in schemas:
            config.components.schemas = {**config.components.schemas, **schemas["schemas"]}
        serializer = Serializer()
        json = serializer.to_json(config)
        document_json = ujson.dumps(
            cls.remove_null(ujson.loads(json)), allow_nan=False,
            indent=4
        )
        document_yml = serializer.to_yaml(config)
        document_obj = serializer.to_obj(config)
        return _Document(document_json, document_yml, document_obj)

    @classmethod
    def setup(cls, path: str, app: "NestipyApplication", config: OpenAPI):
        def register_open_api():
            document: _Document = cls._create_document(app, config)
            api_path = f"/{path.strip('/')}"

            async def openapi_json_handler(_req: Request, res: Response, _next_fn):
                return await res.status(200).header('Content-Type', 'application/json').send(document.json)

            async def openapi_yml_handler(_req: Request, res: Response, _next_fn):
                return await res.status(200).header('Content-Type', 'text/yml').send(document.yml)

            async def openapi_swagger_handler(_req: Request, res: Response, _next_fn):
                file = await aiofiles.open(os.path.join(os.path.dirname(__file__), 'swagger.html'), 'r')
                response = await res.status(200).html(await file.read())
                await file.close()
                return response

            http_adapter = app.get_adapter()
            http_adapter.get('/openapi.json', openapi_json_handler, {})
            http_adapter.get('/openapi.yml', openapi_yml_handler, {})
            http_adapter.get(api_path, openapi_swagger_handler, {})

        Reflect.set_metadata(app, OPENAPI_HANDLER_METADATA, register_open_api)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nestipy.types_.handler import NextFn
    from nestipy.core.context.execution_context import ExecutionContext


class NestipyInterceptor(ABC):

    @abstractmethod
    async def intercept(self, context: "ExecutionContext", next_fn: "NextFn"):
        """
        Args:
            context (ExecutionContext): ExecutionContext of request
            next_fn (NextFn): The next function handler
        """
        pass



import dataclasses
from typing import Any, Optional, Literal, Union

from dataclasses_json import dataclass_json

from .exception import RpcException

MICROSERVICE_CHANNEL = "channel:microservice"


@dataclasses.dataclass
class RpcRequest:
    data: Any
    pattern: Any
    response_topic: Optional[Any] = None
    headers: dict[str, str] = dataclasses.field(default_factory=lambda: {})

    def is_event(self) -> bool:
        return self.response_topic is None


@dataclass_json
@dataclasses.dataclass
class RpcResponse:
    pattern: str
    data: Any
    status: Literal["success", "error"]
    exception: Union[RpcException, None] = None
    headers: dict[str, str] = dataclasses.field(default_factory=lambda: {})
